import subprocess
import psutil
from cnvrg.helpers.logger_helper import log_message
import shlex
import threading
import re
import sys
import os
ON_POSIX = 'posix' in sys.builtin_module_names


def __send_cmd(cmd, cwd=None, env=None):
    env = dict({**os.environ, **(env or {}), **{'PYTHONUNBUFFERED': "1"}})
    cmd = re.sub(r"^(python3?)",r"{exe} -u".format(exe=sys.executable), cmd)
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)
    params = {
        'stdout': subprocess.PIPE,
        'stderr': subprocess.PIPE,
        'close_fds': False,
        'bufsize': 1,
    }
    if cwd:
        params['cwd'] = r"{}".format(cwd)
    if env:
        params["env"] = env
    return subprocess.Popen(cmd, **params)


def __track_logs(buffer, output, print_output):
    for log in buffer:
        log = log.decode("utf-8")
        output.append(log)
        if print_output:
            log_message(log)

def run_sync(cmd, print_output=False, cwd=None):
    proc = __send_cmd(cmd, cwd=cwd)
    output = []
    t1 = threading.Thread(target=__track_logs, args=(proc.stdout, output, print_output,))
    t2 = threading.Thread(target=__track_logs, args=(proc.stderr, output, print_output,))
    t1.start()
    t2.start()
    t1.join()
    t2.join()
    return proc.returncode

def run_async(cmd):
    return __send_cmd(cmd)

def analyze_pid(proc=None, pid=None):
    if proc: pid = proc.pid
    if not pid: return
    p = psutil.Process(pid)
    with p.oneshot():
        return {"cpu": p.cpu_times(), "cpu_precent": p.cpu_percent(), "memory_info": p.memory_percent(), "threads": p.num_threads(), "name": p.name()}


import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.modules.errors as errors
class BaseConnector():
    def __init__(self, data_connector, info=None):
        org, data_connector = param_build_helper.parse_params(data_connector, param_build_helper.DATA_CONNECTOR)
        self._org = org
        self._data_connector = data_connector
        self.info = None
        self.__init_data_connector(info=info)

    @staticmethod
    def key_type():
        raise errors.CnvrgError("Not implemented")

    def connector_type(self):
        return self.info.get("connector_type") or 'dataset'

    def __init_data_connector(self, info=None):
        self.info = info or apis_helper.get(self.__base_url()).get("data_connector")
        self.data = self.info.get("credentials")

    def __base_url(self):
        return apis_helper.url_join('users', self._org, "data_connectors", self._data_connector)

    @classmethod
    def create(cls, title, **kwargs):
        data_connector = {"title": title, "credentials": kwargs, "connector_type": cls.key_type()}
        resp = apis_helper.post(apis_helper.url_join("users", apis_helper.credentials.owner, 'data_connectors'), data={"data_connector": data_connector})
        payload = resp.get("data_connector")
        if not payload: raise errors.CnvrgError("Cant create data connector")
        return cls(payload.get("slug"))


import click
from cnvrg.modules.project import Project

def clone_project(project_url):
    owner_slug, project_slug = Project.get_owner_and_project_from_url(project_url)
    click.secho("Cloning {}".format(project_slug), fg="green")
    project = Project(owner_slug=owner_slug, project_slug=project_slug)
    click.secho(project.clone())



def test_experiment(project_url, command):
    owner_slug, project_slug = Project.get_owner_and_project_from_url(project_url)
    project = Project(owner_slug=owner_slug, project_slug=project_slug)
    exp = project.create_experiment(command, title="Hello Test ME :D")




import json
import yaml
import cnvrg.modules.errors as errors
import os

def load_hyper_search(path_to_hypersearch):
    if isinstance(path_to_hypersearch, dict): return path_to_hypersearch
    path_to_hypersearch = __hyper_path(path_to_hypersearch)
    with open(path_to_hypersearch, 'r') as f:
        if path_to_hypersearch.endswith(".yaml") or path_to_hypersearch.endswith(".yml"):
            return yaml.safe_load(f)
        if path_to_hypersearch.endswith(".json"):
            return json.load(f)
    raise errors.CnvrgError("Cant load hypersearch from {path} hypersearch should be in json or yml format".format(path=path_to_hypersearch))


def __hyper_path(path):
    if not os.path.exists(path):
        path = os.path.join(os.curdir, path)
        if not os.path.exists(path): raise errors.CnvrgError("Cant find hypersearch in {path}".format(path=path))
    return path



import os
import yaml

CONFIG_TYPE_PROJECT = "PROJECT"
CONFIG_TYPE_DATASET = "DATASET"

def find_config_dir(path=None):
    currdir = path or os.curdir
    while os.path.realpath(currdir) != os.path.expanduser("~") and os.path.realpath(currdir) != "/":
        if os.path.exists(os.path.join(currdir, '.cnvrg', 'config.yml')):
            return currdir
        currdir = os.path.join(currdir, '..')
    return False
import pdb
def config_path(path=None):
    config_dir = find_config_dir(path=path)
    if not config_dir: return
    return os.path.join(config_dir, ".cnvrg", "config.yml")

def load_config(path=None):
    if not path: return {}
    with open(config_path(path), 'r') as f:
        return yaml.safe_load(f)

def config_type(path=None):
    if not find_config_dir(path):
        return None
    config = load_config(path=path)
    if "project" in config or ":project_name" in config:
        return CONFIG_TYPE_PROJECT
    return CONFIG_TYPE_DATASET


def save_config(config, path=None):
    config_dir = find_config_dir(path=path)
    config_dir = os.path.join(config_dir or path, ".cnvrg")
    os.makedirs(config_dir, exist_ok=True)
    config_path = os.path.join(config_dir, "config.yml")
    with open(config_path, "w+") as f:
        yaml.dump(config, f)

def is_in_dir(dir_type, dir_element, working_dir):
    working_dir = working_dir or os.curdir
    if not find_config_dir(working_dir): return False
    if not config_type(working_dir) == dir_type: return False
    config = load_config(working_dir)
    return (config.get(":project_name") or config.get("project_name") or config.get("project") or config.get(":dataset_name") or config.get("dataset_name") or config.get("dataset")) == dir_element


def get_element_slug(working_dir=None):
    config = load_config(find_config_dir(working_dir))
    return config.get(":project_name") or config.get("project_name") or config.get("project") or config.get(":dataset_name") or config.get("dataset_name") or config.get("dataset")


def get_element_owner(working_dir=None):
    config = load_config(find_config_dir(working_dir))
    return config.get(":owner") or config.get("owner")





from cnvrg.modules.cnvrg_files import CnvrgFiles
import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.helpers.config_helper as config_helper
from cnvrg.modules.errors import CnvrgError
from cnvrg.modules.storage import storage_factory
import cnvrg.helpers.data_working_dir_helper as data_working_dir_helper
import os.path

class Dataset(CnvrgFiles):
    def __init__(self, dataset=None, dataset_url=None, working_dir=None):
        super(Dataset, self).__init__()
        working_dir = data_working_dir_helper.init_data_working_dir(working_dir)
        if dataset_url:
            owner_slug, project_slug = Dataset.get_owner_and_project_from_url(dataset_url)
        else:
            owner_slug, project_slug = param_build_helper.parse_params(dataset, param_build_helper.DATASET, working_dir=working_dir)
        self.__owner = owner_slug
        self.__dataset = project_slug
        in_dir = config_helper.is_in_dir(config_helper.CONFIG_TYPE_DATASET, project_slug, working_dir)
        working_dir = os.path.join(working_dir, self.__dataset)
        self._set_working_dir(working_dir, save=True)
        self.storage = storage_factory(self, self.get_working_dir())
        if in_dir:
            self._set_working_dir(config_helper.find_config_dir(path=working_dir))
            self.in_dir = True
        elif not self.__dataset or not self.__owner:
            raise CnvrgError("Cant init dataset without params and outside project directory")

    def get_base_url(self):
        return "users/{owner}/datasets/{dataset}".format(owner=self.__owner, dataset=self.__dataset)

    def fetch_all_files(self):
        return self.get_commit_files()

    def download_file(self, file):
        file_path = os.path.join(self.get_working_dir(), file.get("fullpath"))
        if not os.path.exists(file_path):
            self.storage.download_single_file(file)
        return file_path

    # def pytorch_dataset(self, loader=None):
    #     from torch.utils.data.dataset import Dataset as TorchDataset
    #     class CnvrPytorchDataset(TorchDataset):
    #         'Characterizes a dataset for PyTorch'
    #         def __init__(self, dataset: CnvrgFiles=None):
    #             'Initialization'
    #             self.dataset = dataset
    #             self.files = dict(map(lambda x: [x.get("fullpath"), x], dataset.get_commit_files()))
    #             self.keys = list(self.files.keys())
    #             self.storage = storage_factory(dataset)
    #
    #         def __len__(self):
    #             'Denotes the total number of samples'
    #             return len(self.files.keys())
    #
    #         def __getitem__(self, index):
    #             'Generates one sample of data'
    #             # Select sample
    #             file_def = self.files[self.keys[index]]
    #             file_path = os.path.join(self.dataset.get_working_dir(), file_def.get("fullpath"))
    #             if not os.path.exists(file_path):
    #                 self.storage.download_single_file(file_def)
    #             return file_path if not loader else loader(file_path)
    #
    #     return CnvrPytorchDataset(dataset=self)



import json
def parse_args(args):
    return dict(map(lambda x: x.split(":"), args))


def parse_datasets(datasets):
    parsed_datasets = []
    for dataset in datasets:
        commit = None
        dataset_slug = dataset
        if ":" in dataset:
            dataset_slug, commit = dataset.split(":")
        parsed_datasets += [{"slug": dataset_slug, "commit": commit}]
    return parsed_datasets



def __filter_o(o):
    if isinstance(o, list):
        return __remove_nones_array(o)
    if isinstance(o, dict):
        return __remove_nones_dict(o)
    return o

def __filter_bool(o):
    if o == None: return False
    if o == '': return False
    return True

def __remove_nones_dict(o):
    o_keys = filter(lambda x: __filter_bool(o[x]), o.keys())
    return {k: __filter_o(o[k]) for k in o_keys}


def __remove_nones_array(a):
    return list(filter(__filter_bool, [__filter_o(o) for o in a]))

def print_object(o, filter_nones=False):
    if filter_nones: o = __filter_o(o)
    print(json.dumps(o, sort_keys=True, indent=4))